{-# OPTIONS_GHC -fno-warn-unused-binds -fno-warn-missing-signatures #-}
{-# LANGUAGE CPP #-}
{-# LINE 1 "Lexer.x" #-}

{-

Open Source Fixedpoint Model-Checker version 2018

(C) Copyright Sebastian Moedersheim 2003,2018
(C) Copyright Paolo Modesti 2012
(C) Copyright Nicklas Bo Jensen 2012
(C) Copyright IBM Corp. 2009
(C) Copyright ETH Zurich (Swiss Federal Institute of Technology) 2003,2007

All Rights Reserved.

-}

module Lexer (Token(..), 
              Ident, 
	      AlexPosn(..), alexScanTokens, 
	      token_posn) where
import Ast
import Msg



#if __GLASGOW_HASKELL__ >= 603
#include "ghcconfig.h"
#elif defined(__GLASGOW_HASKELL__)
#include "config.h"
#endif
#if __GLASGOW_HASKELL__ >= 503
import Data.Array
import Data.Array.Base (unsafeAt)
#else
import Array
#endif
{-# LINE 1 "templates/wrappers.hs" #-}
-- -----------------------------------------------------------------------------
-- Alex wrapper code.
--
-- This code is in the PUBLIC DOMAIN; you may copy it freely and use
-- it for any purpose whatsoever.






import Data.Word (Word8)
















import Data.Char (ord)
import qualified Data.Bits

-- | Encode a Haskell String to a list of Word8 values, in UTF8 format.
utf8Encode :: Char -> [Word8]
utf8Encode = map fromIntegral . go . ord
 where
  go oc
   | oc <= 0x7f       = [oc]

   | oc <= 0x7ff      = [ 0xc0 + (oc `Data.Bits.shiftR` 6)
                        , 0x80 + oc Data.Bits..&. 0x3f
                        ]

   | oc <= 0xffff     = [ 0xe0 + (oc `Data.Bits.shiftR` 12)
                        , 0x80 + ((oc `Data.Bits.shiftR` 6) Data.Bits..&. 0x3f)
                        , 0x80 + oc Data.Bits..&. 0x3f
                        ]
   | otherwise        = [ 0xf0 + (oc `Data.Bits.shiftR` 18)
                        , 0x80 + ((oc `Data.Bits.shiftR` 12) Data.Bits..&. 0x3f)
                        , 0x80 + ((oc `Data.Bits.shiftR` 6) Data.Bits..&. 0x3f)
                        , 0x80 + oc Data.Bits..&. 0x3f
                        ]



type Byte = Word8

-- -----------------------------------------------------------------------------
-- The input type


type AlexInput = (AlexPosn,     -- current position,
                  Char,         -- previous char
                  [Byte],       -- pending bytes on current char
                  String)       -- current input string

ignorePendingBytes :: AlexInput -> AlexInput
ignorePendingBytes (p,c,_ps,s) = (p,c,[],s)

alexInputPrevChar :: AlexInput -> Char
alexInputPrevChar (_p,c,_bs,_s) = c

alexGetByte :: AlexInput -> Maybe (Byte,AlexInput)
alexGetByte (p,c,(b:bs),s) = Just (b,(p,c,bs,s))
alexGetByte (_,_,[],[]) = Nothing
alexGetByte (p,_,[],(c:s))  = let p' = alexMove p c
                                  (b:bs) = utf8Encode c
                              in p' `seq`  Just (b, (p', c, bs, s))





























































-- -----------------------------------------------------------------------------
-- Token positions

-- `Posn' records the location of a token in the input text.  It has three
-- fields: the address (number of chacaters preceding the token), line number
-- and column of a token within the file. `start_pos' gives the position of the
-- start of the file and `eof_pos' a standard encoding for the end of file.
-- `move_pos' calculates the new position after traversing a given character,
-- assuming the usual eight character tab stops.


data AlexPosn = AlexPn !Int !Int !Int
        deriving (Eq,Show)

alexStartPos :: AlexPosn
alexStartPos = AlexPn 0 1 1

alexMove :: AlexPosn -> Char -> AlexPosn
alexMove (AlexPn a l c) '\t' = AlexPn (a+1)  l     (((c+alex_tab_size-1) `div` alex_tab_size)*alex_tab_size+1)
alexMove (AlexPn a l _) '\n' = AlexPn (a+1) (l+1)   1
alexMove (AlexPn a l c) _    = AlexPn (a+1)  l     (c+1)


-- -----------------------------------------------------------------------------
-- Default monad
















































































































-- -----------------------------------------------------------------------------
-- Monad (with ByteString input)







































































































-- -----------------------------------------------------------------------------
-- Basic wrapper

























-- -----------------------------------------------------------------------------
-- Basic wrapper, ByteString version
































-- -----------------------------------------------------------------------------
-- Posn wrapper

-- Adds text positions to the basic model.


--alexScanTokens :: String -> [token]
alexScanTokens str0 = go (alexStartPos,'\n',[],str0)
  where go inp__@(pos,_,_,str) =
          case alexScan inp__ 0 of
                AlexEOF -> []
                AlexError ((AlexPn _ line column),_,_,_) -> error $ "lexical error at line " ++ (show line) ++ ", column " ++ (show column)
                AlexSkip  inp__' _ln     -> go inp__'
                AlexToken inp__' len act -> act pos (take len str) : go inp__'



-- -----------------------------------------------------------------------------
-- Posn wrapper, ByteString version














-- -----------------------------------------------------------------------------
-- GScan wrapper

-- For compatibility with previous versions of Alex, and because we can.














alex_tab_size :: Int
alex_tab_size = 8
alex_base :: Array Int Int
alex_base = listArray (0 :: Int, 120)
  [ -8
  , -119
  , -38
  , -73
  , 0
  , 40
  , 0
  , 168
  , 296
  , -54
  , -53
  , 552
  , 488
  , 0
  , 3
  , 734
  , 0
  , 0
  , 0
  , 0
  , -114
  , 0
  , 0
  , 0
  , 0
  , -31
  , 0
  , -25
  , -24
  , 0
  , 0
  , 0
  , -42
  , 0
  , 0
  , 0
  , 0
  , 697
  , 942
  , 1017
  , 1092
  , 1167
  , 1242
  , 1317
  , 1392
  , 1467
  , 1542
  , 1617
  , 1692
  , 1767
  , 1842
  , 1917
  , 1992
  , 2067
  , 2142
  , 2217
  , 2292
  , 2367
  , 2442
  , 2517
  , 2592
  , 2667
  , 2742
  , 2817
  , 2892
  , 2967
  , 3042
  , 3117
  , 3192
  , 3267
  , 3342
  , 3417
  , 3492
  , 3567
  , 3642
  , 3717
  , 3792
  , 3867
  , 3942
  , 4017
  , 4092
  , 4167
  , 4242
  , 4317
  , 4392
  , 4467
  , 4542
  , 4617
  , 4692
  , 4767
  , 4842
  , 4917
  , 4992
  , 5067
  , 5142
  , 5217
  , 5292
  , 5367
  , 5442
  , 5517
  , 5592
  , 5667
  , 5742
  , 5817
  , 5892
  , 5967
  , 6042
  , 6117
  , 6192
  , 6267
  , 6342
  , 6417
  , 6492
  , 6567
  , 6642
  , 6717
  , 6792
  , 6867
  , 6942
  , 7017
  , 772
  ]

alex_table :: Array Int Int
alex_table = listArray (0 :: Int, 7272)
  [ 0
  , 14
  , 14
  , 14
  , 14
  , 14
  , 19
  , 10
  , 27
  , 25
  , 18
  , 24
  , 14
  , 14
  , 14
  , 14
  , 14
  , 26
  , 29
  , 31
  , 0
  , 0
  , 0
  , 0
  , 14
  , 32
  , 0
  , 15
  , 0
  , 30
  , 0
  , 28
  , 16
  , 17
  , 2
  , 14
  , 34
  , 9
  , 33
  , 0
  , 120
  , 120
  , 120
  , 120
  , 120
  , 120
  , 120
  , 120
  , 120
  , 120
  , 22
  , 23
  , 0
  , 0
  , 0
  , 0
  , 0
  , 84
  , 92
  , 92
  , 92
  , 92
  , 92
  , 74
  , 92
  , 92
  , 92
  , 97
  , 92
  , 92
  , 92
  , 92
  , 100
  , 92
  , 92
  , 92
  , 72
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 35
  , 0
  , 36
  , 0
  , 0
  , 0
  , 91
  , 87
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 117
  , 92
  , 92
  , 92
  , 81
  , 92
  , 92
  , 92
  , 105
  , 92
  , 92
  , 92
  , 20
  , 1
  , 21
  , 12
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 3
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 8
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 11
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 120
  , 120
  , 120
  , 120
  , 120
  , 120
  , 120
  , 120
  , 120
  , 120
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 11
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 13
  , 12
  , 8
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 4
  , 3
  , 7
  , 6
  , 6
  , 6
  , 5
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 38
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 39
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 40
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 42
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 43
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 44
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 46
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 119
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 49
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 50
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 118
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 54
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 116
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 115
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 56
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 114
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 59
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 113
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 112
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 111
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 62
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 63
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 110
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 109
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 64
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 66
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 108
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 68
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 107
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 69
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 106
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 75
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 77
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 104
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 103
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 89
  , 102
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 79
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 80
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 101
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 83
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 96
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 95
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 93
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 94
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 90
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 98
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 99
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 88
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 86
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 85
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 82
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 78
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 73
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 71
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 70
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 76
  , 92
  , 92
  , 67
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 65
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 61
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 60
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 58
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 57
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 55
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 53
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 52
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 51
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 48
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 47
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 45
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 41
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 92
  , 0
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 37
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 92
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  , 0
  ]

alex_check :: Array Int Int
alex_check = listArray (0 :: Int, 7272)
  [ -1
  , 9
  , 10
  , 11
  , 12
  , 13
  , 125
  , 45
  , 62
  , 62
  , 124
  , 42
  , 9
  , 10
  , 11
  , 12
  , 13
  , 42
  , 42
  , 61
  , -1
  , -1
  , -1
  , -1
  , 32
  , 33
  , -1
  , 35
  , -1
  , 37
  , -1
  , 62
  , 40
  , 41
  , 42
  , 32
  , 44
  , 45
  , 46
  , -1
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , 58
  , 59
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , 91
  , -1
  , 93
  , -1
  , -1
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 123
  , 124
  , 125
  , 191
  , 192
  , 193
  , 194
  , 195
  , 196
  , 197
  , 198
  , 199
  , 200
  , 201
  , 202
  , 203
  , 204
  , 205
  , 206
  , 207
  , 208
  , 209
  , 210
  , 211
  , 212
  , 213
  , 214
  , 215
  , 216
  , 217
  , 218
  , 219
  , 220
  , 221
  , 222
  , 223
  , 224
  , 225
  , 226
  , 227
  , 228
  , 229
  , 230
  , 231
  , 232
  , 233
  , 234
  , 235
  , 236
  , 237
  , 238
  , 239
  , 240
  , 241
  , 242
  , 243
  , 244
  , 245
  , 246
  , 247
  , 248
  , 249
  , 250
  , 251
  , 252
  , 253
  , 254
  , 255
  , 143
  , 144
  , 145
  , 146
  , 147
  , 148
  , 149
  , 150
  , 151
  , 152
  , 153
  , 154
  , 155
  , 156
  , 157
  , 158
  , 159
  , 160
  , 161
  , 162
  , 163
  , 164
  , 165
  , 166
  , 167
  , 168
  , 169
  , 170
  , 171
  , 172
  , 173
  , 174
  , 175
  , 176
  , 177
  , 178
  , 179
  , 180
  , 181
  , 182
  , 183
  , 184
  , 185
  , 186
  , 187
  , 188
  , 189
  , 190
  , 191
  , 192
  , 193
  , 194
  , 195
  , 196
  , 197
  , 198
  , 199
  , 200
  , 201
  , 202
  , 203
  , 204
  , 205
  , 206
  , 207
  , 208
  , 209
  , 210
  , 211
  , 212
  , 213
  , 214
  , 215
  , 216
  , 217
  , 218
  , 219
  , 220
  , 221
  , 222
  , 223
  , 224
  , 225
  , 226
  , 227
  , 228
  , 229
  , 230
  , 231
  , 232
  , 233
  , 234
  , 235
  , 236
  , 237
  , 238
  , 239
  , 240
  , 241
  , 242
  , 243
  , 244
  , 245
  , 246
  , 247
  , 248
  , 249
  , 250
  , 251
  , 252
  , 253
  , 254
  , 255
  , 128
  , 129
  , 130
  , 131
  , 132
  , 133
  , 134
  , 135
  , 136
  , 137
  , 138
  , 139
  , 140
  , 141
  , 142
  , 143
  , 144
  , 145
  , 146
  , 147
  , 148
  , 149
  , 150
  , 151
  , 152
  , 153
  , 154
  , 155
  , 156
  , 157
  , 158
  , 159
  , 160
  , 161
  , 162
  , 163
  , 164
  , 165
  , 166
  , 167
  , 168
  , 169
  , 170
  , 171
  , 172
  , 173
  , 174
  , 175
  , 176
  , 177
  , 178
  , 179
  , 180
  , 181
  , 182
  , 183
  , 184
  , 185
  , 186
  , 187
  , 188
  , 189
  , 190
  , 191
  , 192
  , 193
  , 194
  , 195
  , 196
  , 197
  , 198
  , 199
  , 200
  , 201
  , 202
  , 203
  , 204
  , 205
  , 206
  , 207
  , 208
  , 209
  , 210
  , 211
  , 212
  , 213
  , 214
  , 215
  , 216
  , 217
  , 218
  , 219
  , 220
  , 221
  , 222
  , 223
  , 224
  , 225
  , 226
  , 227
  , 228
  , 229
  , 230
  , 231
  , 232
  , 233
  , 234
  , 235
  , 236
  , 237
  , 238
  , 239
  , 240
  , 241
  , 242
  , 243
  , 244
  , 245
  , 246
  , 247
  , 248
  , 249
  , 250
  , 251
  , 252
  , 253
  , 254
  , 255
  , 128
  , 129
  , 130
  , 131
  , 132
  , 133
  , 134
  , 135
  , 136
  , 137
  , 138
  , 139
  , 140
  , 141
  , 142
  , 143
  , 144
  , 145
  , 146
  , 147
  , 148
  , 149
  , 150
  , 151
  , 152
  , 153
  , 154
  , 155
  , 156
  , 157
  , 158
  , 159
  , 160
  , 161
  , 162
  , 163
  , 164
  , 165
  , 166
  , 167
  , 168
  , 169
  , 170
  , 171
  , 172
  , 173
  , 174
  , 175
  , 176
  , 177
  , 178
  , 179
  , 180
  , 181
  , 182
  , 183
  , 184
  , 185
  , 186
  , 187
  , 188
  , 189
  , 190
  , 191
  , 192
  , 193
  , 194
  , 195
  , 196
  , 197
  , 198
  , 199
  , 200
  , 201
  , 202
  , 203
  , 204
  , 205
  , 206
  , 207
  , 208
  , 209
  , 210
  , 211
  , 212
  , 213
  , 214
  , 215
  , 216
  , 217
  , 218
  , 219
  , 220
  , 221
  , 222
  , 223
  , 224
  , 225
  , 226
  , 227
  , 228
  , 229
  , 230
  , 231
  , 232
  , 233
  , 234
  , 235
  , 236
  , 237
  , 238
  , 239
  , 240
  , 241
  , 242
  , 243
  , 244
  , 245
  , 246
  , 247
  , 248
  , 249
  , 250
  , 251
  , 252
  , 253
  , 254
  , 255
  , 0
  , 1
  , 2
  , 3
  , 4
  , 5
  , 6
  , 7
  , 8
  , 9
  , 10
  , 11
  , 12
  , 13
  , 14
  , 15
  , 16
  , 17
  , 18
  , 19
  , 20
  , 21
  , 22
  , 23
  , 24
  , 25
  , 26
  , 27
  , 28
  , 29
  , 30
  , 31
  , 32
  , 33
  , 34
  , 35
  , 36
  , 37
  , 38
  , 39
  , 40
  , 41
  , 42
  , 43
  , 44
  , 45
  , 46
  , 47
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , 58
  , 59
  , 60
  , 61
  , 62
  , 63
  , 64
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , 91
  , 92
  , 93
  , 94
  , 95
  , 96
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 123
  , 124
  , 125
  , 126
  , 127
  , 192
  , 193
  , 194
  , 195
  , 196
  , 197
  , 198
  , 199
  , 200
  , 201
  , 202
  , 203
  , 204
  , 205
  , 206
  , 207
  , 208
  , 209
  , 210
  , 211
  , 212
  , 213
  , 214
  , 215
  , 216
  , 217
  , 218
  , 219
  , 220
  , 221
  , 222
  , 223
  , 224
  , 225
  , 226
  , 227
  , 228
  , 229
  , 230
  , 231
  , 232
  , 233
  , 234
  , 235
  , 236
  , 237
  , 238
  , 239
  , 240
  , 241
  , 242
  , 243
  , 244
  , 245
  , 246
  , 247
  , 248
  , 249
  , 250
  , 251
  , 252
  , 253
  , 254
  , 255
  , 10
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 128
  , 129
  , 130
  , 131
  , 132
  , 133
  , 134
  , 135
  , 136
  , 137
  , 138
  , 139
  , 140
  , 141
  , 142
  , 143
  , 144
  , 145
  , 146
  , 147
  , 148
  , 149
  , 150
  , 151
  , 152
  , 153
  , 154
  , 155
  , 156
  , 157
  , 158
  , 159
  , 160
  , 161
  , 162
  , 163
  , 164
  , 165
  , 166
  , 167
  , 168
  , 169
  , 170
  , 171
  , 172
  , 173
  , 174
  , 175
  , 176
  , 177
  , 178
  , 179
  , 180
  , 181
  , 182
  , 183
  , 184
  , 185
  , 186
  , 187
  , 188
  , 189
  , 190
  , 191
  , 192
  , 193
  , 194
  , 195
  , 196
  , 197
  , 198
  , 199
  , 200
  , 201
  , 202
  , 203
  , 204
  , 205
  , 206
  , 207
  , 208
  , 209
  , 210
  , 211
  , 212
  , 213
  , 214
  , 215
  , 216
  , 217
  , 218
  , 219
  , 220
  , 221
  , 222
  , 223
  , 224
  , 225
  , 226
  , 227
  , 228
  , 229
  , 230
  , 231
  , 232
  , 233
  , 234
  , 235
  , 236
  , 237
  , 238
  , 239
  , 240
  , 241
  , 242
  , 243
  , 244
  , 245
  , 246
  , 247
  , 248
  , 249
  , 250
  , 251
  , 252
  , 253
  , 254
  , 255
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , 48
  , 49
  , 50
  , 51
  , 52
  , 53
  , 54
  , 55
  , 56
  , 57
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , 65
  , 66
  , 67
  , 68
  , 69
  , 70
  , 71
  , 72
  , 73
  , 74
  , 75
  , 76
  , 77
  , 78
  , 79
  , 80
  , 81
  , 82
  , 83
  , 84
  , 85
  , 86
  , 87
  , 88
  , 89
  , 90
  , -1
  , -1
  , -1
  , -1
  , 95
  , -1
  , 97
  , 98
  , 99
  , 100
  , 101
  , 102
  , 103
  , 104
  , 105
  , 106
  , 107
  , 108
  , 109
  , 110
  , 111
  , 112
  , 113
  , 114
  , 115
  , 116
  , 117
  , 118
  , 119
  , 120
  , 121
  , 122
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  ]

alex_deflt :: Array Int Int
alex_deflt = listArray (0 :: Int, 120)
  [ -1
  , -1
  , -1
  , 13
  , 13
  , 4
  , 4
  , -1
  , -1
  , -1
  , -1
  , 15
  , 15
  , 15
  , -1
  , 15
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  , -1
  ]

alex_accept = listArray (0 :: Int, 120)
  [ AlexAccNone
  , AlexAccNone
  , AlexAccNone
  , AlexAccNone
  , AlexAccNone
  , AlexAccNone
  , AlexAccNone
  , AlexAccNone
  , AlexAccNone
  , AlexAccNone
  , AlexAccNone
  , AlexAccNone
  , AlexAccNone
  , AlexAccNone
  , AlexAccSkip
  , AlexAccSkip
  , AlexAcc 104
  , AlexAcc 103
  , AlexAcc 102
  , AlexAcc 101
  , AlexAcc 100
  , AlexAcc 99
  , AlexAcc 98
  , AlexAcc 97
  , AlexAcc 96
  , AlexAcc 95
  , AlexAcc 94
  , AlexAcc 93
  , AlexAcc 92
  , AlexAcc 91
  , AlexAcc 90
  , AlexAcc 89
  , AlexAcc 88
  , AlexAcc 87
  , AlexAcc 86
  , AlexAcc 85
  , AlexAcc 84
  , AlexAcc 83
  , AlexAcc 82
  , AlexAcc 81
  , AlexAcc 80
  , AlexAcc 79
  , AlexAcc 78
  , AlexAcc 77
  , AlexAcc 76
  , AlexAcc 75
  , AlexAcc 74
  , AlexAcc 73
  , AlexAcc 72
  , AlexAcc 71
  , AlexAcc 70
  , AlexAcc 69
  , AlexAcc 68
  , AlexAcc 67
  , AlexAcc 66
  , AlexAcc 65
  , AlexAcc 64
  , AlexAcc 63
  , AlexAcc 62
  , AlexAcc 61
  , AlexAcc 60
  , AlexAcc 59
  , AlexAcc 58
  , AlexAcc 57
  , AlexAcc 56
  , AlexAcc 55
  , AlexAcc 54
  , AlexAcc 53
  , AlexAcc 52
  , AlexAcc 51
  , AlexAcc 50
  , AlexAcc 49
  , AlexAcc 48
  , AlexAcc 47
  , AlexAcc 46
  , AlexAcc 45
  , AlexAcc 44
  , AlexAcc 43
  , AlexAcc 42
  , AlexAcc 41
  , AlexAcc 40
  , AlexAcc 39
  , AlexAcc 38
  , AlexAcc 37
  , AlexAcc 36
  , AlexAcc 35
  , AlexAcc 34
  , AlexAcc 33
  , AlexAcc 32
  , AlexAcc 31
  , AlexAcc 30
  , AlexAcc 29
  , AlexAcc 28
  , AlexAcc 27
  , AlexAcc 26
  , AlexAcc 25
  , AlexAcc 24
  , AlexAcc 23
  , AlexAcc 22
  , AlexAcc 21
  , AlexAcc 20
  , AlexAcc 19
  , AlexAcc 18
  , AlexAcc 17
  , AlexAcc 16
  , AlexAcc 15
  , AlexAcc 14
  , AlexAcc 13
  , AlexAcc 12
  , AlexAcc 11
  , AlexAcc 10
  , AlexAcc 9
  , AlexAcc 8
  , AlexAcc 7
  , AlexAcc 6
  , AlexAcc 5
  , AlexAcc 4
  , AlexAcc 3
  , AlexAcc 2
  , AlexAcc 1
  , AlexAcc 0
  ]

alex_actions = array (0 :: Int, 105)
  [ (104,alex_action_2)
  , (103,alex_action_3)
  , (102,alex_action_4)
  , (101,alex_action_5)
  , (100,alex_action_6)
  , (99,alex_action_7)
  , (98,alex_action_8)
  , (97,alex_action_9)
  , (96,alex_action_10)
  , (95,alex_action_11)
  , (94,alex_action_12)
  , (93,alex_action_13)
  , (92,alex_action_14)
  , (91,alex_action_15)
  , (90,alex_action_16)
  , (89,alex_action_17)
  , (88,alex_action_18)
  , (87,alex_action_19)
  , (86,alex_action_20)
  , (85,alex_action_21)
  , (84,alex_action_22)
  , (83,alex_action_23)
  , (82,alex_action_24)
  , (81,alex_action_25)
  , (80,alex_action_26)
  , (79,alex_action_27)
  , (78,alex_action_28)
  , (77,alex_action_29)
  , (76,alex_action_30)
  , (75,alex_action_31)
  , (74,alex_action_32)
  , (73,alex_action_33)
  , (72,alex_action_34)
  , (71,alex_action_35)
  , (70,alex_action_35)
  , (69,alex_action_35)
  , (68,alex_action_35)
  , (67,alex_action_35)
  , (66,alex_action_35)
  , (65,alex_action_35)
  , (64,alex_action_35)
  , (63,alex_action_35)
  , (62,alex_action_35)
  , (61,alex_action_35)
  , (60,alex_action_35)
  , (59,alex_action_35)
  , (58,alex_action_35)
  , (57,alex_action_35)
  , (56,alex_action_35)
  , (55,alex_action_35)
  , (54,alex_action_35)
  , (53,alex_action_35)
  , (52,alex_action_35)
  , (51,alex_action_35)
  , (50,alex_action_35)
  , (49,alex_action_35)
  , (48,alex_action_35)
  , (47,alex_action_35)
  , (46,alex_action_35)
  , (45,alex_action_35)
  , (44,alex_action_35)
  , (43,alex_action_35)
  , (42,alex_action_35)
  , (41,alex_action_35)
  , (40,alex_action_35)
  , (39,alex_action_35)
  , (38,alex_action_35)
  , (37,alex_action_35)
  , (36,alex_action_35)
  , (35,alex_action_35)
  , (34,alex_action_35)
  , (33,alex_action_35)
  , (32,alex_action_35)
  , (31,alex_action_35)
  , (30,alex_action_35)
  , (29,alex_action_35)
  , (28,alex_action_35)
  , (27,alex_action_35)
  , (26,alex_action_35)
  , (25,alex_action_35)
  , (24,alex_action_35)
  , (23,alex_action_35)
  , (22,alex_action_35)
  , (21,alex_action_35)
  , (20,alex_action_35)
  , (19,alex_action_35)
  , (18,alex_action_35)
  , (17,alex_action_35)
  , (16,alex_action_35)
  , (15,alex_action_35)
  , (14,alex_action_35)
  , (13,alex_action_35)
  , (12,alex_action_35)
  , (11,alex_action_35)
  , (10,alex_action_35)
  , (9,alex_action_35)
  , (8,alex_action_35)
  , (7,alex_action_35)
  , (6,alex_action_35)
  , (5,alex_action_35)
  , (4,alex_action_35)
  , (3,alex_action_35)
  , (2,alex_action_35)
  , (1,alex_action_35)
  , (0,alex_action_36)
  ]

{-# LINE 74 "Lexer.x" #-}


--- type Ident=String

data Token= 
   TATOM AlexPosn Ident	
   | TOPENP AlexPosn
   | TCLOSEP AlexPosn
   | TOPENSCRYPT AlexPosn
   | TCLOSESCRYPT AlexPosn
   | TOPENB AlexPosn
   | TCLOSEB AlexPosn
   | TCOLON AlexPosn
   | TSEMICOLON AlexPosn
   | TSECCH AlexPosn
   | TAUTHCH AlexPosn
   | TCONFCH AlexPosn
   | TINSECCH AlexPosn
   | TPERCENT AlexPosn
   | TEXCLAM AlexPosn
   | TDOT AlexPosn
   | TCOMMA AlexPosn
   | TOPENSQB AlexPosn
   | TCLOSESQB AlexPosn
   | TPROTOCOL AlexPosn
   | TKNOWLEDGE AlexPosn
   | TTYPES AlexPosn
   | TACTIONS AlexPosn
   | TABSTRACTION AlexPosn
   | TGOALS AlexPosn
   | TFSECCH AlexPosn
   | TFAUTHCH AlexPosn
   | TAUTHENTICATES AlexPosn
   | TON AlexPosn
   | TWEAKLY AlexPosn
   | TSECRET AlexPosn
   | TBETWEEN AlexPosn
   | TUNEQUAL AlexPosn
   | TWHERE AlexPosn
   deriving (Eq,Show)

token_posn (TATOM p _)=p
token_posn (TOPENP p)=p
token_posn (TCLOSEP p)=p
token_posn (TOPENSCRYPT p)=p
token_posn (TCLOSESCRYPT p)=p
token_posn (TOPENB p)=p
token_posn (TCLOSEB p)=p
token_posn (TCOLON p)=p
token_posn (TSEMICOLON p)=p
token_posn (TSECCH p)=p
token_posn (TAUTHCH p)=p
token_posn (TCONFCH p)=p
token_posn (TINSECCH p)=p
token_posn (TPERCENT p)=p
token_posn (TEXCLAM p)=p
token_posn (TDOT p)=p
token_posn (TCOMMA p)=p
token_posn (TOPENSQB p)=p
token_posn (TCLOSESQB p)=p
token_posn (TPROTOCOL p)=p
token_posn (TKNOWLEDGE p)=p
token_posn (TTYPES p)=p
token_posn (TACTIONS p)=p
token_posn (TABSTRACTION p)=p
token_posn (TGOALS p)=p
token_posn (TFSECCH p)=p
token_posn (TFAUTHCH p)=p
token_posn (TAUTHENTICATES p)=p
token_posn (TWEAKLY p)=p
token_posn (TON p)=p
token_posn (TSECRET p)=p
token_posn (TBETWEEN p)=p
token_posn (TWHERE p)=p
token_posn (TUNEQUAL p)=p

alex_action_2 =  (\ p s -> TOPENP p)  
alex_action_3 =  (\ p s -> TCLOSEP p) 
alex_action_4 =  (\ p s -> TOPENSCRYPT p)
alex_action_5 =  (\ p s -> TCLOSESCRYPT p)
alex_action_6 =  (\ p s -> TOPENB p)  
alex_action_7 =  (\ p s -> TCLOSEB p) 
alex_action_8 =  (\ p s -> TCOLON p)  
alex_action_9 =  (\ p s -> TSEMICOLON p) 
alex_action_10 =  (\ p s -> TSECCH p)  
alex_action_11 =  (\ p s -> TAUTHCH p) 
alex_action_12 =  (\ p s -> TCONFCH p) 
alex_action_13 =  (\ p s -> TINSECCH p)
alex_action_14 =  (\ p s -> TFAUTHCH p) 
alex_action_15 =  (\ p s -> TFSECCH p) 
alex_action_16 =  (\ p s -> TPERCENT p)
alex_action_17 =  (\ p s -> TUNEQUAL p)
alex_action_18 =  (\ p s -> TEXCLAM  p)
alex_action_19 =  (\ p s -> TDOT p) 
alex_action_20 =  (\ p s -> TCOMMA p) 
alex_action_21 =  (\ p s -> TOPENSQB p) 
alex_action_22 =  (\ p s -> TCLOSESQB p) 
alex_action_23 =  (\ p s -> TPROTOCOL p) 
alex_action_24 =  (\ p s -> TKNOWLEDGE p)
alex_action_25 =  (\ p s -> TTYPES p)
alex_action_26 =  (\ p s -> TACTIONS p)
alex_action_27 =  (\ p s -> TABSTRACTION p)
alex_action_28 =  (\ p s -> TGOALS p)
alex_action_29 =  (\ p s -> TWHERE p) 
alex_action_30 =  (\ p s -> TAUTHENTICATES p)
alex_action_31 =  (\p s -> TON p)
alex_action_32 =  (\p s -> TWEAKLY p)
alex_action_33 =  (\p s -> TSECRET p)
alex_action_34 =  (\p s -> TBETWEEN p)
alex_action_35 =  (\ p s -> TATOM p s) 
alex_action_36 =  (\ p s -> TATOM p s) 
{-# LINE 1 "templates/GenericTemplate.hs" #-}
-- -----------------------------------------------------------------------------
-- ALEX TEMPLATE
--
-- This code is in the PUBLIC DOMAIN; you may copy it freely and use
-- it for any purpose whatsoever.

-- -----------------------------------------------------------------------------
-- INTERNALS and main scanner engine































































alexIndexInt16OffAddr arr off = arr ! off




















alexIndexInt32OffAddr arr off = arr ! off











quickIndex arr i = arr ! i


-- -----------------------------------------------------------------------------
-- Main lexing routines

data AlexReturn a
  = AlexEOF
  | AlexError  !AlexInput
  | AlexSkip   !AlexInput !Int
  | AlexToken  !AlexInput !Int a

-- alexScan :: AlexInput -> StartCode -> AlexReturn a
alexScan input__ (sc)
  = alexScanUser undefined input__ (sc)

alexScanUser user__ input__ (sc)
  = case alex_scan_tkn user__ input__ (0) input__ sc AlexNone of
  (AlexNone, input__') ->
    case alexGetByte input__ of
      Nothing ->



                                   AlexEOF
      Just _ ->



                                   AlexError input__'

  (AlexLastSkip input__'' len, _) ->



    AlexSkip input__'' len

  (AlexLastAcc k input__''' len, _) ->



    AlexToken input__''' len (alex_actions ! k)


-- Push the input through the DFA, remembering the most recent accepting
-- state it encountered.

alex_scan_tkn user__ orig_input len input__ s last_acc =
  input__ `seq` -- strict in the input
  let
  new_acc = (check_accs (alex_accept `quickIndex` (s)))
  in
  new_acc `seq`
  case alexGetByte input__ of
     Nothing -> (new_acc, input__)
     Just (c, new_input) ->



      case fromIntegral c of { (ord_c) ->
        let
                base   = alexIndexInt32OffAddr alex_base s
                offset = (base + ord_c)
                check  = alexIndexInt16OffAddr alex_check offset

                new_s = if (offset >= (0)) && (check == ord_c)
                          then alexIndexInt16OffAddr alex_table offset
                          else alexIndexInt16OffAddr alex_deflt s
        in
        case new_s of
            (-1) -> (new_acc, input__)
                -- on an error, we want to keep the input *before* the
                -- character that failed, not after.
            _ -> alex_scan_tkn user__ orig_input (if c < 0x80 || c >= 0xC0 then (len + (1)) else len)
                                                -- note that the length is increased ONLY if this is the 1st byte in a char encoding)
                        new_input new_s new_acc
      }
  where
        check_accs (AlexAccNone) = last_acc
        check_accs (AlexAcc a  ) = AlexLastAcc a input__ (len)
        check_accs (AlexAccSkip) = AlexLastSkip  input__ (len)

        check_accs (AlexAccPred a predx rest)
           | predx user__ orig_input (len) input__
           = AlexLastAcc a input__ (len)
           | otherwise
           = check_accs rest
        check_accs (AlexAccSkipPred predx rest)
           | predx user__ orig_input (len) input__
           = AlexLastSkip input__ (len)
           | otherwise
           = check_accs rest


data AlexLastAcc
  = AlexNone
  | AlexLastAcc !Int !AlexInput !Int
  | AlexLastSkip     !AlexInput !Int

data AlexAcc user
  = AlexAccNone
  | AlexAcc Int
  | AlexAccSkip

  | AlexAccPred Int (AlexAccPred user) (AlexAcc user)
  | AlexAccSkipPred (AlexAccPred user) (AlexAcc user)

type AlexAccPred user = user -> AlexInput -> Int -> AlexInput -> Bool

-- -----------------------------------------------------------------------------
-- Predicates on a rule

alexAndPred p1 p2 user__ in1 len in2
  = p1 user__ in1 len in2 && p2 user__ in1 len in2

--alexPrevCharIsPred :: Char -> AlexAccPred _
alexPrevCharIs c _ input__ _ _ = c == alexInputPrevChar input__

alexPrevCharMatches f _ input__ _ _ = f (alexInputPrevChar input__)

--alexPrevCharIsOneOfPred :: Array Char Bool -> AlexAccPred _
alexPrevCharIsOneOf arr _ input__ _ _ = arr ! alexInputPrevChar input__

--alexRightContext :: Int -> AlexAccPred _
alexRightContext (sc) user__ _ _ input__ =
     case alex_scan_tkn user__ input__ (0) input__ sc AlexNone of
          (AlexNone, _) -> False
          _ -> True
        -- TODO: there's no need to find the longest
        -- match when checking the right context, just
        -- the first match will do.

